#!/usr/bin/env python
#-*- coding:utf-8 -*-
#===============================================================================
#  Copyright (C) 2010 Schaefer & Tobies SuC GmbH.
#  Author: Markus Grimm <mgr@schaefer-tobies.de>
#
#  This file is part of Pyontapi.
#
#  Pyontapi is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyontapi is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Pyontapi.  If not, see <http://www.gnu.org/licenses/>.
#
#===============================================================================

""" module for types generation. """

from schtob.pyontapi import errors
from schtob.pyontapi.generator import GENERIC_TYPEDEFS

def indent(text, width=4):
    """ indent ``text`` with ``width`` spaces. """
    
    blanks = width*' '
    lines = text.split('\n')
    if lines:
        return '\n'.join([lines[0]] + [blanks+line for line in lines[1:]])
    return ''

def get_if_set(key_str, value):
    """ if ``value`` is set, return key: value indented by 4 spaces. """
    if not value:
        return ''
    return indent("\n'%s': %s," % (key_str, repr(value)))

class APICommand(object):
    
    """ Representation of API Command for generating API package. """
    
    def __init__(self, name, elements):
        self.name = name
        self.elements = elements
    
    def get_package(self):
        """ get package of this API command. """
        return self.name.split('-')[0]
    
    def get_arguments(self):
        """ Yields input elements. """
        for element in self.elements:
            if not element.is_output:
                yield element
    
    def get_output_fields(self):
        """ Yields output elements. """
        for element in self.elements:
            if element.is_output:
                yield element
    
    def get_py_name(self):
        """ get Python method name for this API command. """
        name = u'_'.join(self.name.split('-')[1:])
        if name == 'break' or name[0].isnumeric():
            name = u'_'.join(self.name.split('-'))
        return name
    
    def get_py_argument_names(self):
        """
        Getter for Python arguments. list contains required arguments at the
        beginning followed by the default arguments.
        """  
        non_default_args = []
        default_args = []
        for arg in self.get_arguments():
            if arg.is_optional:
                default_args.append('%s=%s' % 
                    (arg.name_to_py(), repr(arg.get_default()))
                )
            else:
                non_default_args.append(arg.name_to_py())
        return non_default_args + default_args
    
    def get_py_method(self):
        """ Generate Python method string. """
        arg_classes = [arg.get_py_cls() for arg in self.get_arguments()]
        out_classes = [out.get_py_cls() for out in self.get_output_fields()]
        join_str = ',\n'
        joined_arg_classes = join_str.join(arg_classes)
        joined_out_classes = join_str.join(out_classes)
        if len(out_classes) == 1:
            joined_out_classes += ','
        if len(arg_classes) == 1:
            joined_arg_classes += ','
        
        arguments = ', '.join(['self'] + self.get_py_argument_names())
        
        return (
'''
    def %(name)s(%(arguments)s):
        """ invoke `%(api_name)s`. """
        args = (
            %(joined_arg_classes)s
        )
        fields = (
            %(joined_out_classes)s
        )
        return self._filer.do_api_call(%(api_name)s, args, fields)
'''     
        ) % {
            'name': self.get_py_name(),
            'arguments': arguments,
            'joined_arg_classes': indent(joined_arg_classes, 12),
            'joined_out_classes': indent(joined_out_classes, 12),
            'api_name': repr(self.name),
        }


class TypeDef(object):
    
    """ API Type class. """
    
    fields = (
        'name', 'type_name', 'encrypted', 'nonempty', 'is_optional',
        'is_output', 'is_validated', 'is_array', 
    )
    
    bools = {
        u'true': True,
        u'false': False,
        True: True,
        False: False,
    }
    
    defaults = {
        str: u'',
        bool: False,
    }
    
    def __init__(self, info, var_type):
        self.type           = var_type
        self.name           = info.get(u'name')
        self.encrypted      = info.get(u'encrypted', False)
        self.nonempty       = self.bools[info.get(u'is-nonempty',  u'false')]
        self.is_optional    = self.bools[info.get(u'is-optional',  u'false')]
        self.is_output      = self.bools[info.get(u'is-output',    u'false')]
        self.is_array       = ('[]' in info.get(u'type'))
    
    def get_type(self):
        """
        Get type as string. if type is non generic, this returns  
        types.get_py_cls().
        """
        
        if isinstance(self.type, NamedType):
            return self.type.get_py_cls()
        
        return {
            int: 'int',
            bool: 'bool',
            str: 'str',
        }[self.type]
    
    def name_to_py(self):
        """ Returns a pythonic name for this type. """
        return '_'.join(self.name.split('-'))
    
    def get_default(self):
        """ get default value """
        if self.is_array:
            return None
        return self.defaults.get(self.type, None)
    
    def get_py_type_cls(self):
        """
        Returns the Python code of the type class if the type is used in a 
        complex type. 
        """
        
        return indent(
"""NaTypeElement({
    'name': %(name)s,
    'type': %(type)s,%(encrypted)s%(is-array)s%(nonempty)s%(is-optional)s
})""" % {
          'name'        : repr(self.name),
          'type'        : indent(self.get_type()),
          'encrypted'   : get_if_set('encrypted', self.encrypted),
          'is-array'    : get_if_set('is-array', self.is_array),
          'nonempty'    : get_if_set('nonempty', self.nonempty),
          'is-optional' : get_if_set('is-optional', self.is_optional),
        })
    
    def get_py_cls(self):
        """ Returns the Python code for this type. """
        if self.is_output:
            return (
"""NaField({
    'name': %(name)s,
    'type': %(type)s,%(encrypted)s%(is-array)s
})""") % {
             'name'         : repr(self.name),
             'type'         : indent(self.get_type()),
             'encrypted'    : get_if_set('encrypted', self.encrypted),
             'is-array'     : get_if_set('is-array', self.is_array),
             }
        else:
            return (
"""NaArgument(%(arg_name)s, {
    'name': %(name)s,
    'type': %(type)s,%(encrypted)s%(nonempty)s%(is-optional)s%(is-array)s
})""") % {
             'arg_name'     : self.name_to_py(),
             'name'         : repr(self.name),
             'type'         : indent(self.get_type()),
             'encrypted'    : get_if_set('encrypted', self.encrypted),
             'nonempty'     : get_if_set('nonempty', self.nonempty),
             'is-optional'  : get_if_set('is-optional', self.is_optional),
             'is-array'     : get_if_set('is-array', self.is_array),
        }


class NamedType(object):
    
    """ Class for combined (complex) types. """
    
    def __init__(self, name, elements):
        self.name = name
        self.elements = elements
    
    def get_py_cls(self):
        """ Returns the Python code for this type. """
        elements = [element.get_py_type_cls() for element in self.elements]
        joined_elements = ',\n    '.join(elements)
        return (
"""NaNamedType(%(name)s, [
    %(elements)s,
])""") % {
            'elements': joined_elements,
            'name': repr(self.name),
        }
    
    def duplicate(self):
        """ Clone this type. """
        return NamedType(self.name, self.elements[:])
    
    def set_output_value_for_childs(self, output):
        """ Sets is_output attribute of all child elements to ``output``. """
        for child in self.elements:
            child.is_output = output


def gen_typedefs(api):
    """ Generate typedef classes. """
    typedefs = {}
    try:
        result = api.get_api_list_types()
    except errors.APIFailure, exc:
        raise errors.APIGenerationError(exc.errno,
                'cannot get api types! Error was:\n%s' % exc.reason
        )
    
    for element in result['type-entries']:
        typedefs[element['name']] = NamedType( 
            element['name'], element['type-elements']
        )
    
    for typedef in typedefs.values():
        elements = []   
        for element in typedef.elements:
            var_type = element[u'type'].replace('[]', '')
            if var_type in GENERIC_TYPEDEFS:
                var_type = GENERIC_TYPEDEFS[var_type]
            else:
                var_type = typedefs[var_type]
            _element = TypeDef(element, var_type)
            elements.append(_element)
        typedef.elements = elements
    return typedefs
